//Written by  Kyle Hounslow 2013
// modified by: Ahmad Kaifi, Hassan Althobaiti

// modified again for Robot object tracking by: Kathy Nguyen

//Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software")
//, to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
//and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
//IN THE SOFTWARE.

#define _USE_MATH_DEFINES

#include <sstream>
#include <iostream>
#include <vector>
#include <cmath>
#include "objects.h"
#include <time.h>
#include <stdio.h>
#include "windows.h"
#include <conio.h>
#include "tserial.h"
#include "bot_control.h"

serial comm; //serial is a class type defined in these files, used for referring to the communication device

//initial min and max HSV filter values.
//these will be changed using trackbars
int H_MIN = 0;
int H_MAX = 256;
int S_MIN = 0;
int S_MAX = 256;
int V_MIN = 0;
int V_MAX = 256;
//default capture width and height
const int FRAME_WIDTH = 640;
const int FRAME_HEIGHT = 480;
//max number of objects to be detected in frame
const int MAX_NUM_OBJECTS=50;
//minimum and maximum object area
const int MIN_OBJECT_AREA = 10*10;
const int MAX_OBJECT_AREA = FRAME_HEIGHT*FRAME_WIDTH/1.5;
//names that will appear at the top of each window
const string windowName = "Original Image";
const string windowName1 = "HSV Image";
const string windowName2 = "Thresholded Image";
const string windowName3 = "After Morphological Operations";
const string trackbarWindowName = "Trackbars";

//The following for canny edge detec
Mat dst, detected_edges;
Mat src, src_gray;
int edgeThresh = 1;
int lowThreshold;
int const max_lowThreshold = 100;
int ratio = 3;
int kernel_size = 3;
char* window_name = "Edge Map";

//create color objects so that
//we can use their member functions/information
Object blue("blue"), yellow("yellow"), green("green"), red("red");

//polling time interval
const int NUM_SECONDS = 2;

double rotAngle = 0;

void on_trackbar( int, void* )
{//This function gets called whenever a
    // trackbar position is changed
    
    cout << "H_MIN = " << H_MIN << endl;
    cout << "H_MAX = " << H_MAX << endl;
    cout << "S_MIN = " << S_MIN << endl;
    cout << "S_MAX = " << S_MAX << endl;
    cout << "V_MIN = " << V_MIN << endl;
    cout << "V_MAX = " << V_MAX << endl;
    system("clear");
}

string intToString(int number){
    
    std::stringstream ss;
    ss << number;
    return ss.str();
}


void createTrackbars(){
    //create window for trackbars
    namedWindow(trackbarWindowName,0);
    //create memory to store trackbar name on window
    char TrackbarName[50];
    sprintf( TrackbarName, "H_MIN", H_MIN);
    sprintf( TrackbarName, "H_MAX", H_MAX);
    sprintf( TrackbarName, "S_MIN", S_MIN);
    sprintf( TrackbarName, "S_MAX", S_MAX);
    sprintf( TrackbarName, "V_MIN", V_MIN);
    sprintf( TrackbarName, "V_MAX", V_MAX);
    //create trackbars and insert them into window
    //3 parameters are: the address of the variable that is changing when the trackbar is moved(eg.H_LOW),
    //the max value the trackbar can move (eg. H_HIGH),
    //and the function that is called whenever the trackbar is moved(eg. on_trackbar)
    //                                  ---->    ---->     ---->
    createTrackbar( "H_MIN", trackbarWindowName, &H_MIN, H_MAX, on_trackbar );
    createTrackbar( "H_MAX", trackbarWindowName, &H_MAX, H_MAX, on_trackbar );
    createTrackbar( "S_MIN", trackbarWindowName, &S_MIN, S_MAX, on_trackbar );
    createTrackbar( "S_MAX", trackbarWindowName, &S_MAX, S_MAX, on_trackbar );
    createTrackbar( "V_MIN", trackbarWindowName, &V_MIN, V_MAX, on_trackbar );
    createTrackbar( "V_MAX", trackbarWindowName, &V_MAX, V_MAX, on_trackbar );
}

void drawObject(Object theObjects,Mat &frame, Mat &temp, vector< vector<Point> > contours, vector<Vec4i> hierarchy){
    
    for(int i = 0; i< contours.size(); i++ )
    {
        cv::drawContours(frame,contours,i,theObjects.getColor(),3,8,hierarchy);
        cv::circle(frame,cv::Point(theObjects.getXPos(),theObjects.getYPos()),5,theObjects.getColor());
        cv::putText(frame,intToString(theObjects.getXPos())+ " , " + intToString(theObjects.getYPos()),cv::Point(theObjects.getXPos(),theObjects.getYPos()+20),1,1,theObjects.getColor());
        cv::putText(frame,theObjects.getType(),cv::Point(theObjects.getXPos(),theObjects.getYPos()-20),1,2,theObjects.getColor());
    }
}

void drawObject(int x, int y,Mat &frame){
    
    //use some of the openCV drawing functions to draw crosshairs
    //on your tracked image!
    
    //UPDATE:JUNE 18TH, 2013
    //added 'if' and 'else' statements to prevent
    //memory errors from writing off the screen (ie. (-25,-25) is not within the window!)
    
    circle(frame,Point(x,y),20,Scalar(0,255,0),2);
    if(y-25>0)
        line(frame,Point(x,y),Point(x,y-25),Scalar(0,255,0),2);
    else line(frame,Point(x,y),Point(x,0),Scalar(0,255,0),2);
    if(y+25<FRAME_HEIGHT)
        line(frame,Point(x,y),Point(x,y+25),Scalar(0,255,0),2);
    else line(frame,Point(x,y),Point(x,FRAME_HEIGHT),Scalar(0,255,0),2);
    if(x-25>0)
        line(frame,Point(x,y),Point(x-25,y),Scalar(0,255,0),2);
    else line(frame,Point(x,y),Point(0,y),Scalar(0,255,0),2);
    if(x+25<FRAME_WIDTH)
        line(frame,Point(x,y),Point(x+25,y),Scalar(0,255,0),2);
    else line(frame,Point(x,y),Point(FRAME_WIDTH,y),Scalar(0,255,0),2);
    
    putText(frame,intToString(x)+","+intToString(y),Point(x,y+30),1,1,Scalar(0,255,0),2);
    
}

void morphOps(Mat &thresh){
    
    //create structuring element that will be used to "dilate" and "erode" image.
    //the element chosen here is a 3px by 3px rectangle
    Mat erodeElement = getStructuringElement( MORPH_RECT,Size(3,3));
    //dilate with larger element so make sure object is nicely visible
    Mat dilateElement = getStructuringElement( MORPH_RECT,Size(8,8));
    
    erode(thresh,thresh,erodeElement);
    erode(thresh,thresh,erodeElement);
    
    dilate(thresh,thresh,dilateElement);
    dilate(thresh,thresh,dilateElement);
}

//calibration
void trackFilteredObject(Mat threshold,Mat HSV, Mat &cameraFeed){
    
    int x,y;
    
    Mat temp;
    threshold.copyTo(temp);
    //these two vectors needed for output of findContours
    vector< vector<Point> > contours;
    vector<Vec4i> hierarchy;
    //find contours of filtered image using openCV findContours function
    findContours(temp,contours,hierarchy,CV_RETR_CCOMP,CV_CHAIN_APPROX_SIMPLE );
    //use moments method to find our filtered object
    double refArea = 0;
    bool objectFound = false;
    if (hierarchy.size() > 0) {
        int numObjects = hierarchy.size();
        //if number of objects greater than MAX_NUM_OBJECTS we have a noisy filter
        if(numObjects<MAX_NUM_OBJECTS){
            for (int index = 0; index >= 0; index = hierarchy[index][0]) {
                
                Moments moment = moments((cv::Mat)contours[index]);
                double area = moment.m00;
                
                //if the area is less than 20 px by 20px then it is probably just noise
                //if the area is the same as the 3/2 of the image size, probably just a bad filter
                //we only want the object with the largest area so we safe a reference area each
                //iteration and compare it to the area in the next iteration.
                if(area>MIN_OBJECT_AREA){
                    x = moment.m10/area;
                    y = moment.m01/area;
                    
                    objectFound = true;
                    
                }else objectFound = false;
            }
            
            //let user know you found an object
            if(objectFound ==true){
                //draw object location on screen
                drawObject(x,y,cameraFeed);}
            
        }else putText(cameraFeed,"TOO MUCH NOISE! ADJUST FILTER",Point(0,50),1,2,Scalar(0,0,255),2);
    }
}


//tracking multiple objects
bool trackFilteredObject(Object theObject,Mat threshold,Mat HSV, Mat &cameraFeed){
    
    Mat temp;
    threshold.copyTo(temp);
    //these two vectors needed for output of findContours
    vector< vector<Point> > contours;
    vector<Vec4i> hierarchy;
    //find contours of filtered image using openCV findContours function
    findContours(temp,contours,hierarchy,CV_RETR_CCOMP,CV_CHAIN_APPROX_SIMPLE );
    //use moments method to find our filtered object
    double refArea = 0;
    bool objectFound = false;
    if (hierarchy.size() > 0) {
        int numObjects = hierarchy.size();
        //if number of objects greater than MAX_NUM_OBJECTS we have a noisy filter
        if(numObjects<MAX_NUM_OBJECTS){
            for (int index = 0; index >= 0; index = hierarchy[index][0]) {
                
                Moments moment = moments((cv::Mat)contours[index]);
                double area = moment.m00;
                
                //if the area is less than 20 px by 20px then it is probably just noise
                //if the area is the same as the 3/2 of the image size, probably just a bad filter
                //we only want the object with the largest area so we safe a reference area each
                //iteration and compare it to the area in the next iteration.
                if(area>MIN_OBJECT_AREA){
                
                    theObject.setXPos(moment.m10/area);
                    theObject.setYPos(moment.m01/area);
                    
                    if(theObject.getType() == "blue"){
                        blue.setXPos(theObject.getXPos());
                        blue.setYPos(theObject.getYPos());
                    } else if(theObject.getType() == "yellow"){
                        yellow.setXPos(theObject.getXPos());
                        yellow.setYPos(theObject.getYPos());
                    } else if(theObject.getType() == "red"){
                        red.setXPos(theObject.getXPos());
                        red.setYPos(theObject.getYPos());
                    } else if(theObject.getType() == "green"){
                        green.setXPos(theObject.getXPos());
                        green.setYPos(theObject.getYPos());
                    }

                    //Object.setType(theObject.getType());
                    //Object.setColor(theObject.getColor());
                    
                    //objects.push_back(object);
                    
                    objectFound = true;
                    
                }else objectFound = false;
            }
            //let user know you found an object
            if(objectFound ==true){
                //draw object location on screen
                drawObject(theObject,cameraFeed,temp,contours,hierarchy);
            }
            
        }else putText(cameraFeed,"TOO MUCH NOISE! ADJUST FILTER",Point(0,50),1,2,Scalar(0,0,255),2);
    }
    
    return objectFound;
}

extern double getRot()
{
    return rotAngle;
}
int main(int argc, char* argv[])
{
		/* COM 2 refers to the com port in which the USB to SERIAL port is attached. It is shown by right clicking on my computer, then going to properties and then device manager
9600 is the baud-rate in bits per second */
	comm.startDevice("COM8", 9600);
   
	//while(1){
	//	comm.send_data('0'); //The data is sent through the port
	//	comm.get_data(); //The data is sent through the port
	//	Sleep(500);
	//}
	
	//detection of robot and object
    bool head = false;
    bool back = false;
    bool target = false;
    bool robot = false;
    
    //robot and target coordinates
    int roboY = 0;
    int roboX = 0;
    int targetY = 0;
    int targetX = 0;
    double direction = 0;
    double quadAngle = 0;
    
    //if we would like to calibrate our filter values, set to true.
    bool calibrationMode = false;
    
    //Matrix to store each frame of the webcam feed
    Mat cameraFeed;
    Mat threshold;
    Mat HSV;
    
    if(calibrationMode){
        //create slider bars for HSV filtering
        createTrackbars();
    }
    //video capture object to acquire webcam feed
    VideoCapture capture;
    //open capture object at location zero (default location for webcam)
    capture.open(1);
    
    //set height and width of capture frame
    capture.set(CV_CAP_PROP_FRAME_WIDTH,FRAME_WIDTH);
    capture.set(CV_CAP_PROP_FRAME_HEIGHT,FRAME_HEIGHT);
    //start an infinite loop where webcam feed is copied to cameraFeed matrix
    //all of our operations will be performed within this loop
    waitKey(1000);
    
    clock_t this_time = clock();
    clock_t last_time = this_time;
    double time_counter = 0;
    
    while(1){
        //store image to matrix
        capture.read(cameraFeed);
        
        src = cameraFeed;
        
        if( !src.data )
        { return -1; }
        
        //convert frame from BGR to HSV colorspace
        cvtColor(cameraFeed,HSV,COLOR_BGR2HSV);
        
        if(calibrationMode==true){
            
            //need to find the appropriate color range values
            // calibrationMode must be false
            
            //if in calibration mode, we track objects based on the HSV slider values.
//            cvtColor(cameraFeed,HSV,COLOR_BGR2HSV);
//            inRange(HSV,Scalar(H_MIN,S_MIN,V_MIN),Scalar(H_MAX,S_MAX,V_MAX),threshold);
//            morphOps(threshold);
//            imshow(windowName2,threshold);
//            
//            //the folowing for canny edge detec
//            /// Create a matrix of the same type and size as src (for dst)
//            dst.create( src.size(), src.type() );
//            /// Convert the image to grayscale
//            cvtColor( src, src_gray, CV_BGR2GRAY );
//            /// Create a window
//            namedWindow( window_name, CV_WINDOW_AUTOSIZE );
//            /// Create a Trackbar for user to enter threshold
//            createTrackbar( "Min Threshold:", window_name, &lowThreshold, max_lowThreshold, CannyThreshold );
//            /// Show the image
//            trackFilteredObject(threshold,HSV,cameraFeed);
            
            //if in calibration mode, we track objects based on the HSV slider values.
            cvtColor(cameraFeed,HSV,COLOR_BGR2HSV);
            inRange(HSV,Scalar(H_MIN,S_MIN,V_MIN),Scalar(H_MAX,S_MAX,V_MAX),threshold);
            morphOps(threshold);
            imshow(windowName2,threshold);
            trackFilteredObject(threshold, HSV,cameraFeed);
        }
        else{
            
            //first find blue objects
            cvtColor(cameraFeed,HSV,COLOR_BGR2HSV);
            inRange(HSV,blue.getHSVmin(),blue.getHSVmax(),threshold);
            morphOps(threshold);
            head = trackFilteredObject(blue,threshold,HSV,cameraFeed);
            //then yellows
//            cvtColor(cameraFeed,HSV,COLOR_BGR2HSV);
//            inRange(HSV,yellow.getHSVmin(),yellow.getHSVmax(),threshold);
//            morphOps(threshold);
//            back = trackFilteredObject(yellow,threshold,HSV,cameraFeed);
            //then reds
            cvtColor(cameraFeed,HSV,COLOR_BGR2HSV);
            inRange(HSV,red.getHSVmin(),red.getHSVmax(),threshold);
            morphOps(threshold);
            back = trackFilteredObject(red,threshold,HSV,cameraFeed);
            //then greens
            cvtColor(cameraFeed,HSV,COLOR_BGR2HSV);
            inRange(HSV,green.getHSVmin(),green.getHSVmax(),threshold);
            morphOps(threshold);
            target = trackFilteredObject(green,threshold,HSV,cameraFeed);
            
            //if robot detected, calculate orientation
            if(head == true && back == true){
                robot = true;   //robot found
                //calculate robot center
                roboY = (blue.getYPos() + red.getYPos())/2;
                roboX = (blue.getXPos() + red.getXPos())/2;
                int vectorY = blue.getYPos() - red.getYPos();
                int vectorX = blue.getXPos() - red.getXPos();
                //find orientation of robot using atan2 to get quadrant angle
                quadAngle = (180*atan2(vectorY, vectorX))/M_PI;
            
                //cout << "orientation angle = " << quadAngle << endl;
                //cout << "roboX = " << roboX << endl;
                //cout << "roboY = " << roboY << endl;
                
            }
            
            //if robot and target detected, do calculations for path
            if(robot == true && target == true){
                //get x and y of target
                targetY = green.getYPos();
                targetX = green.getXPos();
                
                //calculate direction for robot to reach target
                direction = (180*atan2(roboY - targetY,roboX - targetX))/M_PI;
                
                //calculate difference fo rotation of robot
                rotAngle = -(quadAngle - direction);
                if(rotAngle < 0)
                    rotAngle += 360;
                
//                cout << "orientation angle = " << quadAngle << endl;
//                cout << "direction angle = " << direction << endl;
                //cout << "rotation = " << rotAngle << endl;
                //cout << endl;
                
                arrowedLine(cameraFeed, Point(roboX, roboY), Point(targetX, targetY), CV_RGB(0, 255, 255), 1, 8, 0, 0.1);
                
                
                //check speed needed for robot to get to target every 2 seconds....
                this_time = clock();
                time_counter += (double)(this_time - last_time);
                last_time = this_time;
                
                if(time_counter > (double)(NUM_SECONDS * CLOCKS_PER_SEC))
                {
                    time_counter -= (double)(NUM_SECONDS * CLOCKS_PER_SEC);
                    
                    //rotation angle thresholds
                    //less than minimum angle makes robot go straight
					if(rotAngle >= 0 && rotAngle < 20|| (rotAngle > 340 && rotAngle <= 360) ){
						cout << "GO STRAIGHT --" << endl;
						comm.send_data('5'); //The data is sent through the port
					}
                    //between min and max angle, turn slowly
                    //else if(rotAngle >= 20 && rotAngle < 90){
                    //    cout << "TURN RIGHT SLOW --" << endl;
					//	comm.send_data('9'); //The data is sent through the port
					//}
					//else if(rotAngle >= 270 && rotAngle < 340){
                    //    cout << "TURN LEFT SLOW --" << endl;
					//	comm.send_data('8'); //The data is sent through the port
					//}
                    //greater or equal to max angle, turn fast
                    else if(rotAngle >= 20 && rotAngle < 180){
                        cout << "TURN RIGHT FAST --" << endl;
						comm.send_data('9'); //The data is sent through the port
					}
					else if(rotAngle <=340 && rotAngle >= 180){
                        cout << "TURN LEFT FAST --" << endl;
						comm.send_data('8'); //The data is sent through the port
					}
                }

                //printf("DebugTime = %f\n", time_counter);
                
            } else{
                //if target is not detected, make robot stop
                //(either target is not detected or robot has reached target area)
                cout << "STOP" << endl;
				comm.send_data('0'); //The data is sent through the port
            }
        }
        
        //show frames 
        //imshow(windowName2,threshold);
        imshow(windowName,cameraFeed);
        //imshow(windowName1,HSV);
        
        //delay 30ms so that screen can refresh.
        //image will not appear without this waitKey() command
        waitKey(30);
    }

	comm.stopDevice(); //The device is closed down
    return 0;
}